from functools import partial
from typing import Optional

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from ._base import FilterAlgorithm
from .util import determine_axes_to_group_by


class Clip(FilterAlgorithm):
    """
    Image clipping filter that clips values below a minimum percentile and above a maximum
    percentile.

    By default, these min and max percentiles are set to 0 and 100 respectively, which will
    result in the filter doing nothing.

    This is a wrapper for :py:func:`numpy.clip` that can optionally linearly expand the dynamic
    range of the data to extend from [0, 1]

    Parameters
    ----------
    p_min : int
        values below this percentile are set to p_min (default 0)
    p_max : int
        values above this percentile are set to p_max (default 100)
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) tiles
    expand_dynamic_range : bool
        If True, linearly expand intensity values to fill [0, 1] after clipping. (default False)
    """

    def __init__(
        self, p_min: int = 0, p_max: int = 100, is_volume: bool = False,
        expand_dynamic_range: bool = False
    ) -> None:

        self.p_min = p_min
        self.p_max = p_max
        self.is_volume = is_volume
        self.expand_dynamic_range = expand_dynamic_range

    _DEFAULT_TESTING_PARAMETERS = {"p_min": 0, "p_max": 100}

    @staticmethod
    def _clip(
        image: xr.DataArray, p_min: int, p_max: int,
        expand_dynamic_range: bool
    ) -> xr.DataArray:
        """Clip values of image"""
        v_min, v_max = np.percentile(image, [p_min, p_max])

        image = image.clip(min=v_min, max=v_max)
        if expand_dynamic_range:
            image /= np.max(image)

        return image

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack. Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        clip = partial(
            self._clip,
            p_min=self.p_min, p_max=self.p_max, expand_dynamic_range=self.expand_dynamic_range
        )
        result = stack.apply(
            clip,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes
        )
        return result



import numpy as np

from starfish.core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from starfish.core.morphology.binary_mask import BinaryMaskCollection
from starfish.core.types import Features
from ._base import AssignTargetsAlgorithm


class Label(AssignTargetsAlgorithm):
    """
    Extract cell ids for features in IntensityTable from a set of segmentation masks.
    """

    def __init__(self, **kwargs) -> None:
        pass

    @classmethod
    def _add_arguments(cls, parser) -> None:
        pass

    @staticmethod
    def _assign(
        masks: BinaryMaskCollection,
        decoded_intensities: DecodedIntensityTable,
        in_place: bool,
    ) -> DecodedIntensityTable:

        cell_ids = (
            Features.AXIS,
            np.full(decoded_intensities.sizes[Features.AXIS], fill_value='nan', dtype='<U8')
        )

        decoded_intensities[Features.CELL_ID] = cell_ids

        for _, mask in masks:
            y_min, y_max = float(mask.y.min()), float(mask.y.max())
            x_min, x_max = float(mask.x.min()), float(mask.x.max())

            in_bbox = decoded_intensities.where(
                (decoded_intensities.y >= y_min)
                & (decoded_intensities.y <= y_max)
                & (decoded_intensities.x >= x_min)
                & (decoded_intensities.x <= x_max),
                drop=True
            )

            in_mask = mask.sel(y=in_bbox.y, x=in_bbox.x)
            spot_ids = in_bbox[Features.SPOT_ID][in_mask.values]
            decoded_intensities[Features.CELL_ID].loc[spot_ids] = mask.name

        return decoded_intensities

    def run(
            self,
            masks: BinaryMaskCollection,
            decoded_intensity_table: DecodedIntensityTable,
            verbose: bool = False,
            in_place: bool = False,
    ) -> DecodedIntensityTable:
        """Extract cell ids for features in IntensityTable from a segmentation label image

        Parameters
        ----------
        masks : BinaryMaskCollection
            binary masks segmenting each cell
        decoded_intensity_table : IntensityTable
            spot information
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)

        Returns
        -------
        IntensityTable :
            IntensityTable with added features variable containing cell ids. Points outside of
            cells will be assigned zero.

        """
        return self._assign(masks, decoded_intensity_table, in_place=in_place)



from dataclasses import dataclass
from typing import Any, Hashable, List, Mapping, MutableMapping, Optional, Tuple

import xarray as xr

from starfish.core.types import Axes, Coordinates, SpotAttributes
from starfish.core.util.logging import Log


AXES_ORDER = (Axes.ROUND, Axes.CH)


@dataclass
class PerImageSliceSpotResults:
    """
    Named tuple that gets returned in every spot finding's image_to_spots method, spot_attrs are the
    SpotAttributes and extras is any extra information collected from the spot finding process.
    """
    spot_attrs: SpotAttributes
    extras: Optional[Mapping[str, Any]]


class SpotFindingResults:
    """
    Wrapper class that describes the results from a spot finding method. The results
    mapping is a collection of (round, ch) indices and their corresponding measured
    SpotAttributes.
    """

    def __init__(
            self,
            imagestack_coords,
            log: Log,
            spot_attributes_list: Optional[List[Tuple]] = None,
    ):
        """
        Construct a SpotFindingResults instance

        Parameters
        -----------
        imagestack_coords : xr.CoordinateArray
            The physical coordinate ranges of the ImageStack spots were found in
        log : Log
            The provenance log information from the ImageStack spots were found in.
        spot_attributes_list : Optional[List[Tuple]]
            If spots were found using ImageStack.transform() the result is a list of
            tuples ((r, channel), PerImageSpotResults). Instantiating SpotFindingResults with
            this list will convert the information to a dictionary.
        """
        spot_attributes_list = spot_attributes_list or []
        self._results: MutableMapping[Tuple, PerImageSliceSpotResults] = {
            tuple(indices[i] for i in AXES_ORDER): spots
            for spots, indices in spot_attributes_list
        }
        self.physical_coord_ranges: Mapping[Hashable, xr.DataArray] = {
            Axes.X.value: imagestack_coords[Coordinates.X.value],
            Axes.Y.value: imagestack_coords[Coordinates.Y.value],
            Axes.ZPLANE.value: imagestack_coords[Coordinates.Z.value]}
        self._log: Log = log

    def __setitem__(self, indices: Mapping[Axes, int], value: PerImageSliceSpotResults):
        """
        Add the round, ch indices and corresponding SpotAttributes to the results dict.

        Parameters
        ----------
        indices: Mapping[Axes, int]
            Mapping of Axes to int values
        spots: SpotAttributes
            Describes spots found on this tile.
        """
        round_ch_index = tuple(indices[i] for i in AXES_ORDER)
        self._results[round_ch_index] = value

    def __getitem__(self, indices: Mapping[Axes, int]) -> PerImageSliceSpotResults:
        """
        Returns the spots found in a given round and ch.

        Parameters
        ----------
        indices: Mapping[Axes, int]
            Mapping of Axes to int values

        Returns
        --------
        SpotAttributes
        """
        round_ch_index = tuple(indices[i] for i in AXES_ORDER)
        return self._results[round_ch_index]

    def items(self):
        """
        Return iterator for (r,ch) and SpotAttributes in Spot finding results
        """
        return self._results.items()

    def keys(self):
        """
        Return all round, ch index pairs.
        """
        return self._results.keys()

    def values(self):
        """
        Return all SpotAttributes across rounds and chs.
        """
        return self._results.values()

    @property
    def round_labels(self):
        """
        Return the set of round labels in the SpotFindingResults
        """
        return sorted(set(r for (r, ch) in self.keys()))

    @property
    def ch_labels(self):
        """
        Return the set of ch labels in the SpotFindingResults
        """

        return sorted(set(ch for (r, ch) in self.keys()))

    def count_total_spots(self):
        """
        Return the total number of unique spots represented in the SpotFindingResults.
        """
        total = 0
        for spot_info in self.values():
            total += spot_info.spot_attrs.data.spot_id.size
        return total

    @property
    def get_physical_coord_ranges(self) -> Mapping[Hashable, xr.DataArray]:
        """
        Returns the physical coordinate ranges the SpotResults cover. Needed
        information for calculating the physical coordinate values of decoded spots.
        """
        return self.physical_coord_ranges

    @property
    def log(self) -> Log:
        """
        Returns a list of pipeline components that have been applied to this get these SpotResults
        as well as their corresponding runtime parameters.

        For more information about provenance logging see
        `Provenance Logging
        <https://spacetx-starfish.readthedocs.io/en/latest/help_and_reference/api/utils/ilogging.html>`_

        Returns
        -------
        Log
        """
        return self._log



import pandas as pd

from starfish.core.types import Coordinates, Features
from ._validated_table import ValidatedTable


class DecodedSpots(ValidatedTable):

    required_fields = {
        Coordinates.X.value,          # spot x-coordinate
        Coordinates.Y.value,          # spot y-coordinate
        Features.TARGET,     # spot gene target
    }

    def __init__(self, decoded_spots: pd.DataFrame) -> None:
        """Construct a decoded_spots instance

        Parameters
        ----------
        decoded_spots : pd.DataFrame

        """
        super().__init__(decoded_spots, DecodedSpots.required_fields)

    def save_csv(self, output_file_name: str) -> None:
        self.data.to_csv(output_file_name, index=False)

    @classmethod
    def load_csv(cls, file_name: str) -> "DecodedSpots":
        return cls(pd.read_csv(file_name))



import json
import sys
from typing import Dict

from slicedimage.backends._base import Backend
from slicedimage.io import resolve_path_or_url

from starfish.core.config import StarfishConfig
from starfish.core.util import click
from .util import (get_schema_path, SpaceTxValidator)


@click.command()
@click.option('--experiment-json', help='image metadata file to validate')
def validate_sptx(experiment_json: str, fuzz: bool=False) -> bool:
    return validate(experiment_json, fuzz)

def validate(experiment_json: str, fuzz: bool=False) -> bool:
    """validate a spaceTx formatted experiment.
    Accepts local filepaths or files hosted at http links.
    Loads configuration from StarfishConfig.

    Parameters
    ----------
    experiment_json : str
        path or URL to a target json object to be validated against the schema passed to this
        object's constructor
    fuzz : bool
        whether or not to perform element-by-element fuzzing.
        If true, will return true and will *not* use warnings.

    Returns
    -------
    bool :
        True, if object valid or fuzz=True, else False

    Examples
    --------
    The following will read the experiment json file provided, downloading it if necessary,
    and begin recursively validating it and all referenced json files (e.g. codebook.json):

        >>> from starfish.core.spacetx_format import validate_sptx
        >>> valid = validate_sptx.validate(json_url)
    """

    config = StarfishConfig()
    valid = True

    # use slicedimage to read the top-level experiment json file passed by the user
    try:
        backend, name, baseurl = resolve_path_or_url(
            experiment_json, backend_config=config.slicedimage)
    except ValueError as exception:
        raise Exception(f"could not load {experiment_json}:\n{exception}")

    with backend.read_contextmanager(name) as fh:
        experiment = json.load(fh)

    # validate experiment.json
    valid &= validate_file(name, "experiment.json", fuzz, backend)

    # loop over all the manifests that are stored in images. Disallowed names will have already been
    # excluded by experiment validation.
    for manifest in experiment['images'].values():
        obj: Dict = dict()
        if not validate_file(manifest, "fov_manifest.json", fuzz, backend, obj):
            valid = False
        else:
            for key, fov in obj['contents'].items():
                valid &= validate_file(fov, 'field_of_view/field_of_view.json', fuzz, backend)

    codebook_file = experiment.get('codebook')
    if codebook_file is not None:
        valid &= validate_file(codebook_file, "codebook/codebook.json", fuzz, backend)

    return valid


def validate_file(file: str, schema: str, fuzz: bool=False,
                  backend: Backend=None, output: Dict=None) -> bool:
    """validate a spaceTx formatted file with a given schema.
    Accepts local filepaths or files hosted at http links.

    Parameters
    ----------
    file : str
        path or URL to a target json object to be validated against the schema passed
    schema : str
        resource path to the schema
    backend : slicedimage.backends._base.Backend
        backend previously loaded from a file path or URL,
        or potentially None if a new backend should be loaded.
    fuzz : bool
        whether or not to perform element-by-element fuzzing.
        If true, will return true and will *not* use warnings.
    output : Dict
        dictionary into which the output object can be stored

    Returns
    -------
    bool :
        True, if object valid or fuzz=True, else False

    Examples
    --------
    The following will read the codebook json file provided, downloading it if necessary:

        >>> from starfish.core.spacetx_format.validate_sptx import validate_file
        >>> valid = validate_sptx.validate_file(json_url, "codebook/codebook.json")
    """

    if backend is None:
        backend, name, baseurl = resolve_path_or_url(file)
    else:
        name = file

    with backend.read_contextmanager(name) as fh:
        obj = json.load(fh)
        if output is not None:
            output.update(obj)

    validator = SpaceTxValidator(get_schema_path(schema, obj))

    if fuzz:
        validator.fuzz_object(obj, file)
        return True
    else:
        return validator.validate_object(obj, file)


if __name__ == "__main__":
    valid = validate_sptx()

    if valid:
        sys.exit(0)
    else:
        sys.exit(1)



from abc import abstractmethod

from starfish.core.morphology.binary_mask import BinaryMaskCollection
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class FilterAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(
            self,
            binary_mask_collection: BinaryMaskCollection,
            *args,
            **kwargs
    ) -> BinaryMaskCollection:
        """Performs a filter on the binary mask collection provided."""
        raise NotImplementedError()



from functools import partial
from typing import Callable, Optional, Tuple, Union

import xarray as xr
from skimage.filters import gaussian

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels, Number
from ._base import FilterAlgorithm
from .util import (
    determine_axes_to_group_by,
    validate_and_broadcast_kernel_size,
)


class GaussianLowPass(FilterAlgorithm):
    """
    Multi-dimensional low-pass gaussian filter. This filter blurs image data, and can be
    useful to apply prior to pixel decoding or watershed segmentation to spread intensity across
    neighboring pixels, accounting for noise that the algorithms are sensitive to.

    This is a thin wrapper around :py:func:`skimage.filters.Gaussian`

    Parameters
    ----------
    sigma : Union[Number, Tuple[Number]]
        Standard deviation for Gaussian kernel.
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.
    """

    def __init__(
            self,
            sigma: Union[Number, Tuple[Number]],
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:

        self.sigma = validate_and_broadcast_kernel_size(sigma, is_volume)
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"sigma": 1}

    @staticmethod
    def _low_pass(
            image: xr.DataArray,
            sigma: Union[Number, Tuple[Number]],
    ) -> xr.DataArray:
        """
        Apply a Gaussian blur operation over a multi-dimensional image.

        Parameters
        ----------
        image : xr.DataArray
            2-d or 3-d image data
        sigma : Union[Number, Tuple[Number]]
            Standard deviation of the Gaussian kernel that will be applied. If a float, an
            isotropic kernel will be assumed, otherwise the dimensions of the kernel give (z, y, x)
         rescale : bool
            If true scales data by max value, if false clips max values to one (default False)

        Returns
        -------
        xr.DataArray :
            Blurred data in same shape as input image

        """

        filtered = gaussian(
            image,
            sigma=sigma, output=None, cval=0, multichannel=False, preserve_range=True, truncate=4.0
        )

        return filtered

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack (default False)
        verbose : bool
            if True, report on filtering progress (default False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        low_pass: Callable = partial(self._low_pass, sigma=self.sigma)
        result = stack.apply(
            low_pass,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            level_method=self.level_method
        )
        return result



import numpy as np
from skimage.morphology import binary_dilation

from starfish.core.morphology.label_image import LabelImage
from starfish.core.types import Axes, Coordinates
from .factories import binary_mask_collection_2d, label_array_2d
from ..binary_mask import BinaryMaskCollection


def test_from_label_image():
    label_image_array, physical_ticks = label_array_2d()

    label_image = LabelImage.from_label_array_and_ticks(
        label_image_array,
        None,
        physical_ticks,
        None,
    )

    mask_collection = BinaryMaskCollection.from_label_image(label_image)
    assert len(mask_collection) == 2

    region_0, region_1 = mask_collection.masks()

    assert region_0.name == '0'
    assert region_1.name == '1'

    assert np.array_equal(region_0, np.ones((1, 6), dtype=np.bool))
    temp = np.ones((2, 3), dtype=np.bool)
    temp[-1, -1] = False
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.Y.value], [0])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.Y.value], [3, 4])
    assert np.array_equal(region_1[Axes.X.value], [3, 4, 5])

    assert np.array_equal(region_0[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_0[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:6])

    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:5])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:6])


def test_uncropped_mask():
    """Test that BinaryMaskCollection.uncropped_mask() works correctly.
    """
    label_image_array, physical_ticks = label_array_2d()

    label_image = LabelImage.from_label_array_and_ticks(
        label_image_array,
        None,
        physical_ticks,
        None,
    )

    mask_collection = BinaryMaskCollection.from_label_image(label_image)
    assert len(mask_collection) == 2

    region_0 = mask_collection.uncropped_mask(0)
    assert region_0.shape == label_image_array.shape
    assert region_0.dtype == np.bool
    assert np.all(region_0[0] == 1)
    assert np.all(region_0[1:5] == 0)

    region_1 = mask_collection.uncropped_mask(1)
    assert region_1.shape == label_image_array.shape
    assert region_1.dtype == np.bool
    assert np.all(region_1[0:3, :] == 0)
    assert np.all(region_1[:, 0:3] == 0)
    assert np.all(region_1[3:5, 3:6] == [[1, 1, 1],
                                         [1, 1, 0]])


def test_uncropped_mask_no_uncropping():
    """If the mask doesn't need to be uncropped, it should still work.  This is an optimized code
    path, so it is separately validated.
    """
    label_image_array, physical_ticks = label_array_2d()
    label_image_array.fill(1)

    label_image = LabelImage.from_label_array_and_ticks(
        label_image_array,
        None,
        physical_ticks,
        None,
    )

    mask_collection = BinaryMaskCollection.from_label_image(label_image)
    assert len(mask_collection) == 1

    region = mask_collection.uncropped_mask(0)
    assert region.shape == label_image_array.shape
    assert np.all(region == 1)


def test_to_label_image():
    # test via roundtrip
    label_image_array, physical_ticks = label_array_2d()

    label_image = LabelImage.from_label_array_and_ticks(
        label_image_array,
        None,
        physical_ticks,
        None,
    )

    masks = BinaryMaskCollection.from_label_image(label_image)

    assert np.array_equal(masks.to_label_image().xarray, label_image.xarray)


def test_save_load(tmp_path):
    binary_mask_collection = binary_mask_collection_2d()

    path = tmp_path / "data.tgz"
    binary_mask_collection.to_targz(path)
    masks2 = BinaryMaskCollection.open_targz(path)
    for m, m2 in zip(binary_mask_collection.masks(), masks2.masks()):
        assert np.array_equal(m, m2)

    # ensure that the regionprops are equal
    for ix in range(len(binary_mask_collection)):
        original_props = binary_mask_collection.mask_regionprops(ix)
        recalculated_props = binary_mask_collection.mask_regionprops(ix)
        assert original_props == recalculated_props


def test_from_empty_label_image(tmp_path):
    label_array, physical_ticks = label_array_2d()
    label_array.fill(0)

    label_image = LabelImage.from_label_array_and_ticks(
        label_array,
        None,
        physical_ticks,
        None,
    )

    binary_mask_collection = BinaryMaskCollection.from_label_image(label_image)
    masks = list(binary_mask_collection.masks())

    assert len(masks) == 0

    path = tmp_path / "data.tgz"
    binary_mask_collection.to_targz(path)
    masks2 = BinaryMaskCollection.open_targz(path)
    for m, m2 in zip(binary_mask_collection.masks(), masks2.masks()):
        assert np.array_equal(m, m2)

    # ensure that the regionprops are equal
    for ix in range(len(binary_mask_collection)):
        original_props = binary_mask_collection.mask_regionprops(ix)
        recalculated_props = binary_mask_collection.mask_regionprops(ix)
        assert original_props == recalculated_props


def test_apply():
    input_mask_collection = binary_mask_collection_2d()
    output_mask_collection = input_mask_collection._apply(binary_dilation)

    assert input_mask_collection._pixel_ticks == output_mask_collection._pixel_ticks
    assert input_mask_collection._physical_ticks == output_mask_collection._physical_ticks
    assert input_mask_collection._log == output_mask_collection._log
    assert len(input_mask_collection) == len(output_mask_collection)

    region_0, region_1 = output_mask_collection.masks()

    assert region_0.name == '0'
    assert region_1.name == '1'

    temp = np.ones((2, 6), dtype=np.bool)
    assert np.array_equal(region_0, temp)
    temp = np.ones((3, 4), dtype=np.bool)
    temp[0, 0] = 0
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.Y.value], [0, 1])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.Y.value], [2, 3, 4])
    assert np.array_equal(region_1[Axes.X.value], [2, 3, 4, 5])



"""
These tests center around creating an ImageStack with labeled indices and verifying that operations
on such an ImageStack work.
"""
import numpy as np

from starfish.core.experiment.builder.test.factories.unique_tiles import unique_data
from starfish.types import Axes
from .factories.unique_tiles import (
    unique_tiles_imagestack, X_COORDS, Y_COORDS, Z_COORDS,
)
from .imagestack_test_utils import verify_physical_coordinates, verify_stack_data
from ..imagestack import ImageStack

ROUND_LABELS = (1, 4, 6)
CH_LABELS = (2, 4, 6, 8)
ZPLANE_LABELS = (3, 4)
HEIGHT = 2
WIDTH = 4

NUM_FOV = 1
NUM_ROUND = len(ROUND_LABELS)
NUM_CH = len(CH_LABELS)
NUM_ZPLANE = len(ZPLANE_LABELS)


def expected_data(round_label: int, ch_label: int, zplane_label: int):
    return unique_data(
        0,
        ROUND_LABELS.index(round_label),
        CH_LABELS.index(ch_label),
        ZPLANE_LABELS.index(zplane_label),
        NUM_FOV, NUM_ROUND, NUM_CH, NUM_ZPLANE,
        HEIGHT, WIDTH)


def setup_imagestack() -> ImageStack:
    return unique_tiles_imagestack(
        ROUND_LABELS, CH_LABELS, ZPLANE_LABELS, HEIGHT, WIDTH)


def test_labeled_indices_read():
    """Build an imagestack with labeled indices (i.e., indices that do not start at 0 or are not
    sequential non-negative integers).  Verify that get_slice behaves correctly.
    """
    stack = setup_imagestack()

    for round_label in stack.axis_labels(Axes.ROUND):
        for ch_label in stack.axis_labels(Axes.CH):
            for zplane_label in stack.axis_labels(Axes.ZPLANE):
                verify_stack_data(
                    stack,
                    {Axes.ROUND: round_label, Axes.CH: ch_label, Axes.ZPLANE: zplane_label},
                    expected_data(round_label, ch_label, zplane_label),
                )


def test_labeled_indices_set_slice():
    """Build an imagestack with labeled indices (i.e., indices that do not start at 0 or are not
    sequential non-negative integers).  Verify that set_slice behaves correctly.
    """
    for round_ in ROUND_LABELS:
        for ch in CH_LABELS:
            for zplane in ZPLANE_LABELS:
                stack = setup_imagestack()
                zeros = np.zeros((HEIGHT, WIDTH), dtype=np.float32)

                stack.set_slice(
                    {Axes.ROUND: round_, Axes.CH: ch, Axes.ZPLANE: zplane}, zeros)

                for selector in stack._iter_axes({Axes.ROUND, Axes.CH, Axes.ZPLANE}):
                    if (selector[Axes.ROUND] == round_
                            and selector[Axes.CH] == ch
                            and selector[Axes.ZPLANE] == zplane):
                        expected_fill_value = zeros
                    else:
                        expected_fill_value = expected_data(
                            selector[Axes.ROUND], selector[Axes.CH], selector[Axes.ZPLANE])

                    verify_stack_data(stack, selector, expected_fill_value)


def test_labeled_indices_sel_single_tile():
    """Select a single tile across each index from an ImageStack with labeled indices.  Verify that
    the data is correct and that the physical coordinates are correctly set."""
    stack = setup_imagestack()

    expected_zplane_coordinates = np.linspace(Z_COORDS[0], Z_COORDS[1], NUM_ZPLANE)

    for selector in stack._iter_axes({Axes.ROUND, Axes.CH, Axes.ZPLANE}):
        subselected = stack.sel(selector)

        # verify that the subselected stack has the correct index labels.
        for index_type in (Axes.ROUND, Axes.CH, Axes.ZPLANE):
            assert subselected.axis_labels(index_type) == [selector[index_type]]

        # verify that the subselected stack has the correct data.
        expected_tile_data = expected_data(
            selector[Axes.ROUND], selector[Axes.CH], selector[Axes.ZPLANE])
        verify_stack_data(stack, selector, expected_tile_data)

        zplane_index = ZPLANE_LABELS.index(selector[Axes.ZPLANE])
        expected_zplane_coordinate = expected_zplane_coordinates[zplane_index]

        # assert that the physical coordinate values are what we expect.
        verify_physical_coordinates(
            stack,
            X_COORDS,
            Y_COORDS,
            expected_zplane_coordinate,
            selector[Axes.ZPLANE],
        )


def test_labeled_indices_sel_slice():
    """Select a single tile across each index from an ImageStack with labeled indices.  Verify that
    the data is correct and that the physical coordinates are correctly set."""
    stack = setup_imagestack()
    set_selector = {Axes.ROUND: slice(None, 4), Axes.CH: slice(4, 6), Axes.ZPLANE: 4}
    subselected = stack.sel(set_selector)

    # verify that the subselected stack has the correct index labels.
    for index_type, expected_results in (
            (Axes.ROUND, [1, 4]),
            (Axes.CH, [4, 6]),
            (Axes.ZPLANE, [4],)):
        assert subselected.axis_labels(index_type) == expected_results

    expected_zplane_coordinates = np.linspace(Z_COORDS[0], Z_COORDS[1], NUM_ZPLANE)

    for selector in subselected._iter_axes({Axes.ROUND, Axes.CH, Axes.ZPLANE}):
        # verify that the subselected stack has the correct data.
        expected_tile_data = expected_data(
            selector[Axes.ROUND], selector[Axes.CH], selector[Axes.ZPLANE])
        verify_stack_data(subselected, selector, expected_tile_data)

        zplane_index = ZPLANE_LABELS.index(set_selector[Axes.ZPLANE])
        expected_zplane_coordinate = expected_zplane_coordinates[zplane_index]

        # verify that each tile in the subselected stack has the correct physical coordinates.
        verify_physical_coordinates(
            stack,
            X_COORDS,
            Y_COORDS,
            expected_zplane_coordinate,
            selector[Axes.ZPLANE]
        )


def multiply(array, value):
    return array * value


def test_labeled_indices_apply():
    """Build an imagestack with labeled indices (i.e., indices that do not start at 0 or are not
    sequential non-negative integers).  Verify that apply behaves correctly.
    """
    stack = setup_imagestack()

    for round_ in stack.axis_labels(Axes.ROUND):
        for ch in stack.axis_labels(Axes.CH):
            for zplane in stack.axis_labels(Axes.ZPLANE):
                verify_stack_data(
                    stack,
                    {Axes.ROUND: round_, Axes.CH: ch, Axes.ZPLANE: zplane},
                    expected_data(round_, ch, zplane),
                )

    output = stack.apply(multiply, value=0.5)

    for round_ in stack.axis_labels(Axes.ROUND):
        for ch in stack.axis_labels(Axes.CH):
            for zplane in stack.axis_labels(Axes.ZPLANE):
                verify_stack_data(
                    output,
                    {Axes.ROUND: round_, Axes.CH: ch, Axes.ZPLANE: zplane},
                    expected_data(round_, ch, zplane) * 0.5,
                )



# deal with numpy import warnings due to cython
# See: https://stackoverflow.com/questions/40845304/
#      runtimewarning-numpy-dtype-size-changed-may-indicate-binary-incompatibility)
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")  # noqa
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")  # noqa

# generated version number and commit hash
# TODO: (ttung) the type: ignore is required to satisfy the new semantic analyzer in mypy 0.710.
from ._version import get_versions  # type: ignore


# NOTE: if we move to python 3.7, we can produce this value at call time via __getattr__
version = get_versions()['version']
is_release_tag = None
if "+" not in str(version):
    is_release_tag = f"Release: {version}"



import numpy as np
import pytest

from ..expand import fill_from_mask


def test_expand_mask_2d(shape=(15, 15)):
    mask = np.array(
        [[True, False, False],
         [False, True, True]],
        dtype=np.bool)
    label_image = np.random.randint(30, np.iinfo(np.uint16).max, size=shape, dtype=np.uint16)
    label_image_clone = np.copy(label_image)
    fill_from_mask(mask, (2, 3), 20, label_image)

    assert label_image.dtype == np.uint16
    assert label_image.shape == shape

    compare_mask = np.ones(shape=shape, dtype=np.bool)
    for y, x in (
            (2, 3),
            (3, 4),
            (3, 5),
    ):
        assert label_image[y, x] == 20
        compare_mask[y, x] = False

    assert np.all(label_image_clone & compare_mask == label_image & compare_mask)


def test_expand_mask_3d(shape=(15, 15, 15)):
    mask = np.array(
        [[[True, False, True],
          [True, False, True],
          [False, False, False],
          [False, True, False]],

         [[True, False, True],
          [False, False, False],
          [True, False, False],
          [False, False, False]]],
        dtype=np.bool)
    label_image = np.random.randint(30, np.iinfo(np.uint16).max, size=shape, dtype=np.uint16)
    label_image_clone = np.copy(label_image)
    fill_from_mask(mask, (2, 7, 3), 20, label_image)

    assert label_image.dtype == np.uint16
    assert label_image.shape == shape

    compare_mask = np.ones(shape=shape, dtype=np.bool)
    for z, y, x in (
            (2, 7, 3),
            (2, 7, 5),
            (2, 8, 3),
            (2, 8, 5),
            (2, 10, 4),
            (3, 7, 3),
            (3, 7, 5),
            (3, 9, 3),
    ):
        assert label_image[z, y, x] == 20
        compare_mask[z, y, x] = False

    assert np.all(label_image_clone & compare_mask == label_image & compare_mask)


def test_expand_image_too_small(shape=(2, 2)):
    mask = np.array(
        [[True, False, False],
         [False, True, True]],
        dtype=np.bool)
    label_image = np.random.randint(30, np.iinfo(np.uint16).max, size=shape, dtype=np.uint16)
    with pytest.raises(ValueError):
        fill_from_mask(mask, (2, 3), 20, label_image)


def test_expand_image_negative_axis_labels(shape=(15, 15)):
    mask = np.array(
        [[True, False, False],
         [False, True, True]],
        dtype=np.bool)
    label_image = np.random.randint(30, np.iinfo(np.uint16).max, size=shape, dtype=np.uint16)
    with pytest.raises(ValueError):
        fill_from_mask(mask, (2, -1), 20, label_image)



import numpy as np

from starfish.core.morphology.binary_mask.test.factories import binary_mask_collection_2d
from starfish.core.types import Axes, FunctionSource
from ..map import Map


def test_apply():
    input_mask_collection = binary_mask_collection_2d()
    filt = Map("morphology.binary_dilation", module=FunctionSource.skimage)
    output_mask_collection = filt.run(input_mask_collection)

    assert input_mask_collection._pixel_ticks == output_mask_collection._pixel_ticks
    assert input_mask_collection._physical_ticks == output_mask_collection._physical_ticks
    assert input_mask_collection._log == output_mask_collection._log
    assert len(input_mask_collection) == len(output_mask_collection)

    region_0, region_1 = output_mask_collection.masks()

    assert region_0.name == '0'
    assert region_1.name == '1'

    temp = np.ones((2, 6), dtype=np.bool)
    assert np.array_equal(region_0, temp)
    temp = np.ones((3, 4), dtype=np.bool)
    temp[0, 0] = 0
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.Y.value], [0, 1])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.Y.value], [2, 3, 4])
    assert np.array_equal(region_1[Axes.X.value], [2, 3, 4, 5])



import numpy as np
import pytest

from starfish import ImageStack
from starfish.core.imagestack.test.factories import unique_tiles_imagestack
from starfish.core.test.factories import (
    two_spot_informative_blank_coded_data_factory,
    two_spot_one_hot_coded_data_factory,
    two_spot_sparse_coded_data_factory,
)
from starfish.types import Axes, FunctionSource
from .._base import FindSpotsAlgorithm
from ..blob import BlobDetector
from ..local_max_peak_finder import LocalMaxPeakFinder
from ..trackpy_local_max_peak_finder import TrackpyLocalMaxPeakFinder

# verify all spot finders handle different coding types
_, ONE_HOT_IMAGESTACK, ONE_HOT_MAX_INTENSITY = two_spot_one_hot_coded_data_factory()
_, SPARSE_IMAGESTACK, SPARSE_MAX_INTENSITY = two_spot_sparse_coded_data_factory()
_, BLANK_IMAGESTACK, BLANK_MAX_INTENSITY = two_spot_informative_blank_coded_data_factory()

# make sure that all spot finders handle empty arrays
EMPTY_IMAGESTACK = ImageStack.from_numpy(np.zeros((4, 2, 10, 100, 100), dtype=np.float32))


def simple_gaussian_spot_detector() -> BlobDetector:
    """create a basic gaussian spot detector"""
    return BlobDetector(
        min_sigma=1,
        max_sigma=4,
        num_sigma=5,
        threshold=0,
        measurement_type='max')


def simple_trackpy_local_max_spot_detector() -> TrackpyLocalMaxPeakFinder:
    """create a basic local max peak finder"""
    return TrackpyLocalMaxPeakFinder(
        spot_diameter=3,
        min_mass=0.01,
        max_size=10,
        separation=2,
        radius_is_gyration=True,
        is_volume=True
    )


def simple_local_max_spot_detector() -> LocalMaxPeakFinder:
    return LocalMaxPeakFinder(
        min_distance=6,
        stringency=0,
        min_obj_area=0,
        max_obj_area=np.inf,
        threshold=0
    )

# initialize spot detectors
gaussian_spot_detector = simple_gaussian_spot_detector()
trackpy_local_max_spot_detector = simple_trackpy_local_max_spot_detector()
local_max_spot_detector = simple_local_max_spot_detector()

# test parameterization
test_parameters = (
    'data_stack, spot_detector',
    [
        (ONE_HOT_IMAGESTACK, gaussian_spot_detector),
        (ONE_HOT_IMAGESTACK, trackpy_local_max_spot_detector),
        (ONE_HOT_IMAGESTACK, local_max_spot_detector),
        (SPARSE_IMAGESTACK, gaussian_spot_detector),
        (SPARSE_IMAGESTACK, trackpy_local_max_spot_detector),
        (SPARSE_IMAGESTACK, local_max_spot_detector),
        (BLANK_IMAGESTACK, gaussian_spot_detector),
        (BLANK_IMAGESTACK, trackpy_local_max_spot_detector),
        (BLANK_IMAGESTACK, local_max_spot_detector),
    ]
)


@pytest.mark.parametrize(*test_parameters)
def test_spot_detection_with_reference_image(
        data_stack: ImageStack,
        spot_detector: FindSpotsAlgorithm,
):
    """This testing method uses a reference image to identify spot locations. Each method should
    find 2 spots in the reference image then measure the intensity of those locations in each
    r/ch pair. The final spot results should represent 2 spots for each r/ch totalling
    2*num_rounds*num_ch spots """

    reference_image = data_stack.reduce((Axes.ROUND, Axes.CH), func=FunctionSource.np("max"))
    spots_results = spot_detector.run(image_stack=data_stack, reference_image=reference_image)
    assert spots_results.count_total_spots() == (2 * data_stack.num_chs * data_stack.num_rounds), \
        "wrong number of spots detected"


@pytest.mark.parametrize(*test_parameters)
def test_spot_detection_no_reference_image(
        data_stack: ImageStack,
        spot_detector: FindSpotsAlgorithm,
):
    """
    This testing method does not provide a reference image, and should therefore check for spots
    in each (round, ch) combination in sequence. The final spot results should represent the
    total number of original spots in the ImageStack.
    """

    spots_results = spot_detector.run(image_stack=data_stack)
    assert spots_results.count_total_spots() == 4, "wrong number of spots detected"

    spots = spot_detector.run(image_stack=EMPTY_IMAGESTACK)
    assert spots.count_total_spots() == 0


def _make_labeled_image() -> ImageStack:
    ROUND_LABELS = (1, 4, 6)
    CH_LABELS = (2, 4, 6, 8)
    ZPLANE_LABELS = (3, 4)
    HEIGHT = 2
    WIDTH = 4

    return unique_tiles_imagestack(
        ROUND_LABELS, CH_LABELS, ZPLANE_LABELS, HEIGHT, WIDTH)


def test_reference_image_spot_detection_with_image_with_labeled_axes():
    """This testing method uses a reference image to identify spot locations."""
    data_stack = _make_labeled_image()
    reference_image = data_stack.reduce((Axes.ROUND, Axes.CH), func=FunctionSource.np("max"))
    spot_results = gaussian_spot_detector.run(image_stack=data_stack,
                                              reference_image=reference_image)
    return spot_results


def test_spot_detection_with_image_with_labeled_axes():
    """This testing method uses no reference image to identify spot locations."""
    data_stack = _make_labeled_image()
    spot_results = gaussian_spot_detector.run(image_stack=data_stack)
    return spot_results



import os
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    Hashable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import xarray as xr
from skimage.measure import regionprops
from skimage.measure._regionprops import _RegionProperties

from starfish.core.morphology.label_image import LabelImage
from starfish.core.morphology.util import (
    _get_axes_names,
    _normalize_physical_ticks,
    _normalize_pixel_ticks,
)
from starfish.core.types import ArrayLike, Axes, Coordinates, Number
from starfish.core.util.logging import Log
from .expand import fill_from_mask


@dataclass
class MaskData:
    binary_mask: np.ndarray
    offsets: Tuple[int, ...]
    region_properties: Optional[_RegionProperties]


class BinaryMaskCollection:
    """Collection of binary masks with a dict-like access pattern.

    Parameters
    ----------
    pixel_ticks : Union[Mapping[Axes, ArrayLike[int]], Mapping[str, ArrayLike[int]]]
        A map from the axis to the values for that axis.
    physical_ticks : Union[Mapping[Coordinates, ArrayLike[Number]], Mapping[str, ArrayLike[Number]]
        A map from the physical coordinate type to the values for axis.  For 2D label images,
        X and Y physical coordinates must be provided.  For 3D label images, Z physical
        coordinates must also be provided.
    masks : Sequence[MaskData]
        A sequence of data for binary masks.

    Attributes
    ----------
    max_shape : Mapping[Axes, int]
        Maximum index of contained masks.
    """
    def __init__(
            self,
            pixel_ticks: Union[Mapping[Axes, ArrayLike[int]], Mapping[str, ArrayLike[int]]],
            physical_ticks: Union[Mapping[Coordinates, ArrayLike[Number]],
                                  Mapping[str, ArrayLike[Number]]],
            masks: Sequence[MaskData],
            log: Optional[Log],
    ):
        self._pixel_ticks: Mapping[Axes, ArrayLike[int]] = _normalize_pixel_ticks(pixel_ticks)
        self._physical_ticks: Mapping[Coordinates, ArrayLike[Number]] = _normalize_physical_ticks(
            physical_ticks)
        self._masks: MutableMapping[int, MaskData] = {}
        self._log: Log = log or Log()

        for ix, mask_data in enumerate(masks):
            if mask_data.binary_mask.ndim not in (2, 3):
                raise TypeError(f"expected 2 or 3 dimensions; got {mask_data.binary_mask.ndim}")
            if mask_data.binary_mask.dtype != np.bool:
                raise ValueError(f"expected dtype of bool; got {mask_data.binary_mask.dtype}")

            self._masks[ix] = mask_data

        if len(self._pixel_ticks) != len(self._physical_ticks):
            raise ValueError(
                "pixel_ticks should have the same cardinality as physical_ticks")
        for axis, coord in zip(*_get_axes_names(len(self._pixel_ticks))):
            if axis not in self._pixel_ticks:
                raise ValueError(f"pixel ticks missing {axis.value} data")
            if coord not in self._physical_ticks:
                raise ValueError(f"physical coordinate ticks missing {coord.value} data")
            if len(self._pixel_ticks[axis]) != len(self._physical_ticks[coord]):
                raise ValueError(
                    f"pixel ticks for {axis.name} does not have the same cardinality as physical "
                    f"coordinates ticks for {coord.name}")

    def __getitem__(self, index: int) -> xr.DataArray:
        return self._format_mask_as_xarray(index)

    def __iter__(self) -> Iterator[Tuple[int, xr.DataArray]]:
        for mask_index in self._masks.keys():
            yield mask_index, self._format_mask_as_xarray(mask_index)

    def __len__(self) -> int:
        return len(self._masks)

    def _format_mask_as_xarray(self, index: int) -> xr.DataArray:
        """Convert a np-based mask into an xarray DataArray."""
        mask_data = self._masks[index]
        max_mask_name_len = len(str(len(self._masks) - 1))

        xr_dims: MutableSequence[str] = []
        xr_coords: MutableMapping[Hashable, Any] = {}

        for ix, (axis, coord) in enumerate(zip(*_get_axes_names(len(self._pixel_ticks)))):
            xr_dims.append(axis.value)
            start_offset = mask_data.offsets[ix]
            end_offset = mask_data.offsets[ix] + mask_data.binary_mask.shape[ix]
            xr_coords[axis.value] = self._pixel_ticks[axis.value][start_offset:end_offset]
            xr_coords[coord.value] = (
                axis.value, self._physical_ticks[coord.value][start_offset:end_offset])

        return xr.DataArray(
            mask_data.binary_mask,
            dims=xr_dims,
            coords=xr_coords,
            name=f"{index:0{max_mask_name_len}d}"
        )

    def uncropped_mask(self, index: int) -> xr.DataArray:
        """Some of the binary mask collections builders will crop the binary masks when constructing
        the collection.  The purpose is to exclude the regions of the image that are entirely False.
        Use this method to obtain the mask sized according to the pixel and physical shape provided
        for the entire binary mask collection, use this method."""
        mask_data = self._masks[index]
        uncropped_shape = tuple(
            len(self._pixel_ticks[axis])
            for axis, _ in zip(*_get_axes_names(len(self._pixel_ticks)))
        )

        if uncropped_shape == mask_data.binary_mask.shape:
            return self._format_mask_as_xarray(index)

        max_mask_name_len = len(str(len(self._masks) - 1))

        xr_dims: MutableSequence[str] = []
        xr_coords: MutableMapping[Hashable, Any] = {}

        for ix, (axis, coord) in enumerate(zip(*_get_axes_names(len(self._pixel_ticks)))):
            xr_dims.append(axis.value)
            xr_coords[axis.value] = self._pixel_ticks[axis.value]
            xr_coords[coord.value] = (axis.value, self._physical_ticks[coord.value])

        image = np.zeros(
            shape=tuple(
                len(self._pixel_ticks[axis])
                for axis, _ in zip(*_get_axes_names(len(self._pixel_ticks)))
            ),
            dtype=np.bool,
        )
        fill_from_mask(
            mask_data.binary_mask,
            mask_data.offsets,
            1,
            image,
        )
        return xr.DataArray(
            image,
            dims=xr_dims,
            coords=xr_coords,
            name=f"{index:0{max_mask_name_len}d}"
        )

    def masks(self) -> Iterator[xr.DataArray]:
        for mask_index in self._masks.keys():
            yield self._format_mask_as_xarray(mask_index)

    def mask_regionprops(self, mask_id: int) -> _RegionProperties:
        """
        Return the region properties for a given mask.

        Parameters
        ----------
        mask_id : int
            The mask ID for the mask.

        Returns
        -------
        _RegionProperties
            The region properties for that mask.
        """
        mask_data = self._masks[mask_id]
        if mask_data.region_properties is None:
            # recreate the label image (but with just this mask)
            image = np.zeros(
                shape=tuple(
                    len(self._pixel_ticks[axis])
                    for axis, _ in zip(*_get_axes_names(len(self._pixel_ticks)))
                ),
                dtype=np.uint32,
            )
            fill_from_mask(
                mask_data.binary_mask,
                mask_data.offsets,
                mask_id + 1,
                image,
            )
            mask_data.region_properties = regionprops(image.data)
        return mask_data.region_properties

    @property
    def max_shape(self) -> Mapping[Axes, int]:
        return {
            axis: len(self._pixel_ticks[axis])
            for ix, (axis, _) in enumerate(zip(*_get_axes_names(len(self._pixel_ticks))))
        }

    @property
    def log(self) -> Log:
        return self._log

    @classmethod
    def from_label_image(cls, label_image: LabelImage) -> "BinaryMaskCollection":
        """Creates binary masks from a label image.  Masks are cropped to the smallest size that
        contains the non-zero values, but pixel and physical coordinates ticks are retained.  Masks
        extracted from BinaryMaskCollections will be cropped.  To extract masks sized to the
        original label image, use :py:meth:`starfish.BinaryMaskCollection.uncropped_mask`.

        Parameters
        ----------
        label_image : LabelImage
            LabelImage to extract binary masks from.

        Returns
        -------
        masks : BinaryMaskCollection
            Masks generated from the label image.
        """
        props = regionprops(label_image.xarray.data)

        pixel_ticks = {
            axis.value: label_image.xarray.coords[axis.value]
            for axis, _ in zip(*_get_axes_names(label_image.xarray.ndim))
            if axis.value in label_image.xarray.coords
        }
        physical_ticks = {
            coord.value: label_image.xarray.coords[coord.value]
            for _, coord in zip(*_get_axes_names(label_image.xarray.ndim))
            if coord.value in label_image.xarray.coords
        }
        masks: Sequence[MaskData] = [
            MaskData(prop.image, prop.bbox[:label_image.xarray.ndim], prop)
            for prop in props
        ]
        log = deepcopy(label_image.log)

        return cls(
            pixel_ticks,
            physical_ticks,
            masks,
            log,
        )

    @classmethod
    def from_binary_arrays_and_ticks(
            cls,
            arrays: Sequence[np.ndarray],
            pixel_ticks: Optional[Union[
                Mapping[Axes, ArrayLike[int]],
                Mapping[str, ArrayLike[int]]]],
            physical_ticks: Union[
                Mapping[Coordinates, ArrayLike[Number]],
                Mapping[str, ArrayLike[Number]]],
            log: Optional[Log],
    ) -> "BinaryMaskCollection":
        """Constructs a BinaryMaskCollection from an array containing the labels, a set of physical
        coordinates, and an optional log of how this label image came to be.  Masks are cropped to
        the smallest size that contains the non-zero values, but pixel and physical coordinates
        ticks are retained.  Masks extracted from BinaryMaskCollections will be cropped.  To extract
        masks sized to the original label image, use
        :py:meth:`starfish.morphology.BinaryMaskCollection.uncropped_mask`.

        Parameters
        ----------
        arrays : Sequence[np.ndarray]
            A set of 2D or 3D binary arrays.  The ordering of the axes must be Y, X for 2D images
            and ZPLANE, Y, X for 3D images.  The arrays must have identical sizes and match the
            sizes of pixel_ticks and physical_ticks.
        pixel_ticks : Optional[Union[Mapping[Axes, ArrayLike[int]], Mapping[str, ArrayLike[int]]]]
            A map from the axis to the values for that axis.  For any axis that exist in the array
            but not in pixel_ticks, the pixel coordinates are assigned from 0..N-1, where N is
            the size along that axis.
        physical_ticks : Union[Mapping[Coordinates, ArrayLike[Number]], Mapping[str,
        ArrayLike[Number]]]
            A map from the physical coordinate type to the values for axis.  For 2D label images,
            X and Y physical coordinates must be provided.  For 3D label images, Z physical
            coordinates must also be provided.
        log : Optional[Log]
            A log of how this label image came to be.

        Returns
        -------
        masks : BinaryMaskCollection
            Masks generated from the label image.
        """
        array_shapes = set(array.shape for array in arrays)
        array_dtypes = set(array.dtype for array in arrays)
        if len(array_shapes) > 1:
            raise ValueError("all masks must be identically sized")
        for array_dtype in array_dtypes:
            if array_dtype != np.bool:
                raise TypeError("arrays must be binary data")

        # normalize the pixel coordinates to Mapping[Axes, ArrayLike[int]]
        pixel_ticks = _normalize_pixel_ticks(pixel_ticks)
        # normalize the physical coordinates to Mapping[Coordinates, ArrayLike[Number]]
        physical_ticks = _normalize_physical_ticks(physical_ticks)

        # If we have 1 or more arrays, verify that the physical coordinate array's size matches the
        # array's size.  In BinaryMaskCollection's constructor, it is verified that the pixel
        # tick array's size matches that of the physical coordinate array's size.
        if len(array_shapes) == 1:
            array_shape = next(iter(array_shapes))

            # verify that we don't have extra
            expected_axes, expected_physical_coords = _get_axes_names(len(array_shape))
            actual_real_coordinates_provided = set(expected_physical_coords)
            actual_real_coordinates_provided.intersection_update(set(physical_ticks.keys()))

            if len(array_shape) != len(actual_real_coordinates_provided):
                raise ValueError(
                    f"physical coordinates provided for {len(actual_real_coordinates_provided)} "
                    f"axes ({actual_real_coordinates_provided}), but the data has "
                    f"{len(array_shape)} axes"
                )

            for ix, (axis, coord) in enumerate(zip(*_get_axes_names(len(array_shape)))):
                if (coord in physical_ticks
                        and len(physical_ticks[coord]) != array_shape[ix]):
                    raise ValueError(
                        f"The size of the array for physical coordinates for {coord.name} "
                        f"({len(physical_ticks[coord])} does not match the array's shape "
                        f"({array_shape[ix]})."
                    )

        # Add all pixel ticks not present.
        for axis, coord in zip(*_get_axes_names(3)):
            if coord in physical_ticks and axis not in pixel_ticks:
                pixel_ticks[axis] = np.arange(0, len(physical_ticks[coord]))

        masks: MutableSequence[MaskData] = list()
        for array in arrays:
            selection_range: Tuple[slice, ...] = BinaryMaskCollection._crop_mask(array)

            masks.append(MaskData(
                array[selection_range],
                tuple(selection.start for selection in selection_range),
                None
            ))

        log = deepcopy(log)

        return cls(
            pixel_ticks,
            physical_ticks,
            masks,
            log,
        )

    @staticmethod
    def _crop_mask(array: np.ndarray) -> Tuple[slice, ...]:
        selection_range: MutableSequence[slice] = list()

        # calculate the first and last True pixel across each axis.
        for axis_num in range(array.ndim):
            projection_axes = tuple(dim for dim in range(array.ndim) if dim != axis_num)
            # max project across the projections
            max_projection = np.amax(array, axis=projection_axes)
            # get the indices of the non-zero elements.  the [0] is because np.nonzero captures
            # the indices of each dimension separately, and we want the one and only dimension.
            non_zero_indices = np.nonzero(max_projection)[0]

            if len(non_zero_indices) == 0:
                selection_range.append(slice(0, 0))
            else:
                selection_range.append(slice(non_zero_indices[0], non_zero_indices[-1] + 1))

        return tuple(selection_range)

    @classmethod
    def from_label_array_and_ticks(
            cls,
            array: np.ndarray,
            pixel_ticks: Optional[Union[
                Mapping[Axes, ArrayLike[int]],
                Mapping[str, ArrayLike[int]]]],
            physical_ticks: Union[
                Mapping[Coordinates, ArrayLike[Number]],
                Mapping[str, ArrayLike[Number]]],
            log: Optional[Log],
    ) -> "BinaryMaskCollection":
        """Constructs a BinaryMaskCollection from an array containing the labels, a set of physical
        coordinates, and an optional log of how this label image came to be.  Masks are cropped to
        the smallest size that contains the non-zero values, but pixel and physical coordinates
        ticks are retained.  Masks extracted from BinaryMaskCollections will be cropped.  To extract
        masks sized to the original label image, use
        :py:meth:`starfish.morphology.BinaryMaskCollection.uncropped_mask`.

        Parameters
        ----------
        array : np.ndarray
            A 2D or 3D array containing the labels.  The ordering of the axes must be Y, X for 2D
            images and ZPLANE, Y, X for 3D images.
        pixel_ticks : Optional[Union[Mapping[Axes, ArrayLike[int]], Mapping[str, ArrayLike[int]]]]
            A map from the axis to the values for that axis.  For any axis that exist in the array
            but not in pixel_ticks, the pixel coordinates are assigned from 0..N-1, where N is
            the size along that axis.
        physical_ticks : Union[Mapping[Coordinates, ArrayLike[Number]], Mapping[str,
        ArrayLike[Number]]]
            A map from the physical coordinate type to the values for axis.  For 2D label images,
            X and Y physical coordinates must be provided.  For 3D label images, Z physical
            coordinates must also be provided.
        log : Optional[Log]
            A log of how this label image came to be.

        Returns
        -------
        masks : BinaryMaskCollection
            Masks generated from the label image.
        """
        # normalize the pixel coordinates to Mapping[Axes, ArrayLike[int]]
        pixel_ticks = _normalize_pixel_ticks(pixel_ticks)
        # normalize the physical coordinates to Mapping[Coordinates, ArrayLike[Number]]
        physical_ticks = _normalize_physical_ticks(physical_ticks)

        for ix, (axis, coord) in enumerate(zip(*_get_axes_names(array.ndim))):
            # We verify that the physical coordinate array's size matches the array's size.  In
            # BinaryMaskCollection's constructor, it is verified that the pixel coordinate array's
            # size matches that of the physical coordinate array's size.
            if coord not in physical_ticks:
                raise ValueError(f"Missing physical coordinates for {coord.name}")
            if len(physical_ticks[coord]) != array.shape[ix]:
                raise ValueError(
                    f"The size of the array for physical coordinates for {coord.name} "
                    f"({len(physical_ticks[coord])} does not match the array's shape "
                    f"({array.shape[ix]})."
                )
            if axis not in pixel_ticks:
                pixel_ticks[axis] = np.arange(0, array.shape[ix])

        props = regionprops(array)

        masks: Sequence[MaskData] = [
            MaskData(prop.image, prop.bbox[:array.ndim], prop)
            for prop in props
        ]
        log = deepcopy(log)

        return cls(
            pixel_ticks,
            physical_ticks,
            masks,
            log,
        )

    def to_label_image(self) -> LabelImage:
        shape = tuple(
            len(self._pixel_ticks[axis])
            for axis in (Axes.ZPLANE, Axes.Y, Axes.X)
            if axis in self._pixel_ticks
        )
        label_image_array = np.zeros(shape=shape, dtype=np.uint16)
        for ix, mask_data in self._masks.items():
            fill_from_mask(
                mask_data.binary_mask,
                mask_data.offsets,
                ix + 1,
                label_image_array,
            )

        return LabelImage.from_label_array_and_ticks(
            label_image_array,
            self._pixel_ticks,
            self._physical_ticks,
            self._log,
        )

    @classmethod
    def open_targz(cls, path: Union[str, Path]) -> "BinaryMaskCollection":
        """Load the collection saved as a .tar.gz file from disk

        Parameters
        ----------
        path : Union[str, Path]
            Path of the tar file to instantiate from.

        Returns
        -------
        masks : BinaryMaskCollection
            Collection of binary masks.
        """
        with open(os.fspath(path), "rb") as fh:
            return _io.BinaryMaskIO.read_versioned_binary_mask(fh)

    def to_targz(self, path: Union[str, Path]):
        """Save the binary masks to disk as a .tar.gz file.

        Parameters
        ----------
        path : Union[str, Path]
            Path of the tar file to write to.
        """
        with open(os.fspath(path), "wb") as fh:
            _io.BinaryMaskIO.write_versioned_binary_mask(fh, self)

    def _apply(
            self,
            function: Callable[[np.ndarray], np.ndarray],
            *args,
            n_processes: Optional[int] = None,
            **kwargs
    ) -> "BinaryMaskCollection":
        """Given a function that takes an ndarray and outputs another, apply that method to all the
        masks in this collection to form a new collection.  All the masks are uncropped before being
        passed to the function.

        Parameters
        ----------
        function : Callable[[np.ndarray], np.ndarray]
            A function that should produce a mask array when given a mask array.
        n_processes : Optional[int]
            The number of processes to use for apply. If None, uses the output of os.cpu_count()
            (default = None).
        """
        if n_processes is None:
            n_processes = os.cpu_count()

        applied_func = partial(
            BinaryMaskCollection._apply_single_mask,
            function=function,
            mask_collection=self,
            args=args,
            kwargs=kwargs,
        )
        selectors = range(len(self._masks))

        with ThreadPoolExecutor(max_workers=n_processes) as tpe:
            results: Iterator[MaskData] = tpe.map(applied_func, selectors)

        return BinaryMaskCollection(
            self._pixel_ticks,
            self._physical_ticks,
            list(results),
            self._log,
        )

    @staticmethod
    def _apply_single_mask(
            mask_index: int,
            mask_collection: "BinaryMaskCollection",
            args: Sequence,
            kwargs: Mapping,
            function: Callable[[np.ndarray], np.ndarray],
    ) -> MaskData:
        """Given a mask collection, and an index, retrieve a mask and apply a function to that mask.
        Return the output along with the offsets of the original mask.  If the original mask is
        uncropped, then the offsets should all be 0.  If the original mask is not uncropped, it is
        propagated from the input masks's offsets.
        """
        input_mask = mask_collection.uncropped_mask(mask_index)
        output_mask = function(input_mask.values, *args, **kwargs)  # type: ignore

        selection_range: Sequence[slice] = BinaryMaskCollection._crop_mask(output_mask)

        return MaskData(
            output_mask[selection_range],
            tuple(selection.start for selection in selection_range),
            None
        )


# these need to be at the end to avoid recursive imports
from . import _io  # noqa



from ._base import DecodeSpotsAlgorithm
from .metric_decoder import MetricDistance
from .per_round_max_channel_decoder import PerRoundMaxChannel
from .simple_lookup_decoder import SimpleLookupDecoder

# autodoc's automodule directive only captures the modules explicitly listed in __all__.
all_filters = {
    filter_name: filter_cls
    for filter_name, filter_cls in locals().items()
    if isinstance(filter_cls, type) and issubclass(filter_cls, DecodeSpotsAlgorithm)
}
__all__ = list(all_filters.keys())



from functools import partial
from typing import Callable, Optional, Tuple, Union

import xarray as xr

from starfish.core.image.Filter.gaussian_low_pass import GaussianLowPass
from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels, Number
from starfish.core.util.levels import levels
from ._base import FilterAlgorithm
from .util import (
    determine_axes_to_group_by,
    validate_and_broadcast_kernel_size,
)


class GaussianHighPass(FilterAlgorithm):
    """
    Applies a Gaussian high pass filter to the ImageStack. This is useful to remove cellular
    autofluorescence, which is typically low frequency.

    This filter works by subtracting a Gaussian low pass filter filtered version of the input image
    from the input image itself. The Gaussian low pass filter is defined in :py:func:`scipy.ndimage`
    and used by :py:func:`skimage.filters.gaussian`.

    Parameters
    ----------
    sigma : Union[Number, Tuple[Number]]
        standard deviation of gaussian kernel
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.
    """

    def __init__(
            self,
            sigma: Union[Number, Tuple[Number]],
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:

        self.sigma = validate_and_broadcast_kernel_size(sigma, is_volume)
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"sigma": 3}

    @staticmethod
    def _high_pass(
            image: xr.DataArray,
            sigma: Union[Number, Tuple[Number]],
    ) -> xr.DataArray:
        """
        Applies a gaussian high pass filter to an image

        Parameters
        ----------
        image : Union[xr.DataArray, np.ndarray]
            2-d or 3-d image data
        sigma : Union[Number, Tuple[Number]]
            Standard deviation of gaussian kernel
        rescale : bool
            If true scales data by max value, if false clips max values to one

        Returns
        -------
        np.ndarray :
            filtered image of the same type and shape as the input image
        """

        blurred = GaussianLowPass._low_pass(image, sigma)
        blurred = levels(blurred)  # clip negative values to 0.
        filtered = image - blurred

        return filtered

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.  Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        high_pass: Callable = partial(self._high_pass, sigma=self.sigma)
        result = stack.apply(
            high_pass,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            level_method=self.level_method
        )
        return result



import skimage
from packaging import version

if version.parse(skimage.__version__) > version.parse("0.14.2"):
    import skimage.transform
    match_histograms = skimage.transform.match_histograms
else:

    """
    Unless otherwise specified by LICENSE.txt files in individual
    directories, all code is

    Copyright (C) 2011, the scikit-image team
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in
        the documentation and/or other materials provided with the
        distribution.
    3. Neither the name of skimage nor the names of its contributors may be
        used to endorse or promote products derived from this software without
        specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS`` AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
    STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
    IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    """

    import numpy as np

    def _match_cumulative_cdf(source, template):
        """
        Return modified source array so that the cumulative density function of
        its values matches the cumulative density function of the template.
        """
        src_values, src_unique_indices, src_counts = np.unique(
            source.ravel(), return_inverse=True, return_counts=True)
        tmpl_values, tmpl_counts = np.unique(template.ravel(), return_counts=True)

        # calculate normalized quantiles for each array
        src_quantiles = np.cumsum(src_counts) / source.size
        tmpl_quantiles = np.cumsum(tmpl_counts) / template.size

        interp_a_values = np.interp(src_quantiles, tmpl_quantiles, tmpl_values)
        return interp_a_values[src_unique_indices].reshape(source.shape)

    def match_histograms(image, reference, multichannel=False):
        """
        Adjust an image so that its cumulative histogram matches that of another.
        The adjustment is applied separately for each channel.

        Parameters
        ----------
        image : ndarray
            Input image. Can be gray-scale or in color.
        reference : ndarray
            Image to match histogram of. Must have the same number of channels as
            image.
        multichannel : bool, optional
            Apply the matching separately for each channel.

        Returns
        -------
        matched : ndarray
            Transformed input image.

        Raises
        ------
        ValueError
            Thrown when the number of channels in the input image and the reference
            differ.

        References
        ----------
        .. [1] http://paulbourke.net/miscellaneous/equalisation/
        """
        if image.ndim != reference.ndim:
            raise ValueError('Image and reference must have the same number of channels.')

        if multichannel:
            if image.shape[-1] != reference.shape[-1]:
                raise ValueError('Number of channels in the input image and reference '
                                 'image must match!')

            matched = np.empty(image.shape, dtype=image.dtype)
            for channel in range(image.shape[-1]):
                matched_channel = _match_cumulative_cdf(
                    image[..., channel], reference[..., channel])
                matched[..., channel] = matched_channel
        else:
            matched = _match_cumulative_cdf(image, reference)

        return matched

if version.parse(skimage.__version__) > version.parse("0.14.2"):
    import skimage.feature
    blob_log = skimage.feature.blob_log
    blob_dog = skimage.feature.blob_dog
else:
    from skimage.feature.blob import _prune_blobs
    import numpy as np
    from scipy.ndimage import gaussian_filter, gaussian_laplace
    from skimage import img_as_float
    from skimage.feature import peak_local_max

    def blob_dog(image, min_sigma=1, max_sigma=50, sigma_ratio=1.6, threshold=2.0,
                 overlap=.5, *, exclude_border=False):
        """
        Finds blobs in the given grayscale image.
        Blobs are found using the Difference of Gaussian (DoG) method [1]_.
        For each blob found, the method returns its coordinates and the standard
        deviation of the Gaussian kernel that detected the blob.

        Parameters
        ----------
        image : 2D or 3D ndarray
            Input grayscale image, blobs are assumed to be light on dark
            background (white on black).
        min_sigma : scalar or sequence of scalars, optional
            The minimum standard deviation for Gaussian kernel. Keep this low to
            detect smaller blobs. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number, in
            which case it is equal for all axes.
        max_sigma : scalar or sequence of scalars, optional
            The maximum standard deviation for Gaussian kernel. Keep this high to
            detect larger blobs. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number, in
            which case it is equal for all axes.
        sigma_ratio : float, optional
            The ratio between the standard deviation of Gaussian Kernels used for
            computing the Difference of Gaussians
        threshold : float, optional.
            The absolute lower bound for scale space maxima. Local maxima smaller
            than thresh are ignored. Reduce this to detect blobs with less
            intensities.
        overlap : float, optional
            A value between 0 and 1. If the area of two blobs overlaps by a
            fraction greater than `threshold`, the smaller blob is eliminated.
        exclude_border : int or bool, optional
            If nonzero int, `exclude_border` excludes blobs from
            within `exclude_border`-pixels of the border of the image.

        Returns
        -------
        A : (n, image.ndim + sigma) ndarray
            A 2d array with each row representing 2 coordinate values for a 2D
            image, and 3 coordinate values for a 3D image, plus the sigma(s) used.
            When a single sigma is passed, outputs are:
            ``(r, c, sigma)`` or ``(p, r, c, sigma)`` where ``(r, c)`` or
            ``(p, r, c)`` are coordinates of the blob and ``sigma`` is the standard
            deviation of the Gaussian kernel which detected the blob. When an
            anisotropic gaussian is used (sigmas per dimension), the detected sigma
            is returned for each dimension.

        References
        ----------
        [1] https://en.wikipedia.org/wiki/Blob_detection#The_difference_of_Gaussians_approach

        Examples
        --------
        >>> from skimage import data, feature
        >>> feature.blob_dog(data.coins(), threshold=.5, max_sigma=40)
        array([[ 267.      ,  359.      ,   16.777216],
            [ 267.      ,  115.      ,   10.48576 ],
            [ 263.      ,  302.      ,   16.777216],
            [ 263.      ,  245.      ,   16.777216],
            [ 261.      ,  173.      ,   16.777216],
            [ 260.      ,   46.      ,   16.777216],
            [ 198.      ,  155.      ,   10.48576 ],
            [ 196.      ,   43.      ,   10.48576 ],
            [ 195.      ,  102.      ,   16.777216],
            [ 194.      ,  277.      ,   16.777216],
            [ 193.      ,  213.      ,   16.777216],
            [ 185.      ,  347.      ,   16.777216],
            [ 128.      ,  154.      ,   10.48576 ],
            [ 127.      ,  102.      ,   10.48576 ],
            [ 125.      ,  208.      ,   10.48576 ],
            [ 125.      ,   45.      ,   16.777216],
            [ 124.      ,  337.      ,   10.48576 ],
            [ 120.      ,  272.      ,   16.777216],
            [  58.      ,  100.      ,   10.48576 ],
            [  54.      ,  276.      ,   10.48576 ],
            [  54.      ,   42.      ,   16.777216],
            [  52.      ,  216.      ,   16.777216],
            [  52.      ,  155.      ,   16.777216],
            [  45.      ,  336.      ,   16.777216]])

        Notes
        -----
        The radius of each blob is approximately :math:`sqrt{2}sigma` for
        a 2-D image and :math:`sqrt{3}sigma` for a 3-D image.
        """

        image = img_as_float(image)

        # if both min and max sigma are scalar, function returns only one sigma
        scalar_sigma = np.isscalar(max_sigma) and np.isscalar(min_sigma)

        # Gaussian filter requires that sequence-type sigmas have same
        # dimensionality as image. This broadcasts scalar kernels
        if np.isscalar(max_sigma):
            max_sigma = np.full(image.ndim, max_sigma, dtype=float)
        if np.isscalar(min_sigma):
            min_sigma = np.full(image.ndim, min_sigma, dtype=float)

        # Convert sequence types to array
        min_sigma = np.asarray(min_sigma, dtype=float)
        max_sigma = np.asarray(max_sigma, dtype=float)

        # k such that min_sigma*(sigma_ratio**k) > max_sigma
        k = int(np.mean(np.log(max_sigma / min_sigma) / np.log(sigma_ratio) + 1))

        # a geometric progression of standard deviations for gaussian kernels
        sigma_list = np.array([min_sigma * (sigma_ratio ** i)
                              for i in range(k + 1)])

        gaussian_images = [gaussian_filter(image, s) for s in sigma_list]

        # computing difference between two successive Gaussian blurred images
        # multiplying with average standard deviation provides scale invariance
        dog_images = [(gaussian_images[i] - gaussian_images[i + 1])
                      * np.mean(sigma_list[i]) for i in range(k)]

        image_cube = np.stack(dog_images, axis=-1)

        # local_maxima = get_local_maxima(image_cube, threshold)
        local_maxima = peak_local_max(image_cube, threshold_abs=threshold,
                                      footprint=np.ones((3,) * (image.ndim + 1)),
                                      threshold_rel=0.0,
                                      exclude_border=exclude_border)
        # Catch no peaks
        if local_maxima.size == 0:
            return np.empty((0, 3))

        # Convert local_maxima to float64
        lm = local_maxima.astype(np.float64)

        # translate final column of lm, which contains the index of the
        # sigma that produced the maximum intensity value, into the sigma
        sigmas_of_peaks = sigma_list[local_maxima[:, -1]]

        if scalar_sigma:
            # select one sigma column, keeping dimension
            sigmas_of_peaks = sigmas_of_peaks[:, 0:1]

        # Remove sigma index and replace with sigmas
        lm = np.hstack([lm[:, :-1], sigmas_of_peaks])

        return _prune_blobs(lm, overlap)

    def blob_log(image, min_sigma=1, max_sigma=50, num_sigma=10, threshold=.2,
                 overlap=.5, log_scale=False, *, exclude_border=False):
        """
        Finds blobs in the given grayscale image.
        Blobs are found using the Laplacian of Gaussian (LoG) method [1]_.
        For each blob found, the method returns its coordinates and the standard
        deviation of the Gaussian kernel that detected the blob.

        Parameters
        ----------
        image : 2D or 3D ndarray
            Input grayscale image, blobs are assumed to be light on dark
            background (white on black).
        min_sigma : scalar or sequence of scalars, optional
            the minimum standard deviation for Gaussian kernel. Keep this low to
            detect smaller blobs. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number, in
            which case it is equal for all axes.
        max_sigma : scalar or sequence of scalars, optional
            The maximum standard deviation for Gaussian kernel. Keep this high to
            detect larger blobs. The standard deviations of the Gaussian filter
            are given for each axis as a sequence, or as a single number, in
            which case it is equal for all axes.
        num_sigma : int, optional
            The number of intermediate values of standard deviations to consider
            between `min_sigma` and `max_sigma`.
        threshold : float, optional.
            The absolute lower bound for scale space maxima. Local maxima smaller
            than thresh are ignored. Reduce this to detect blobs with less
            intensities.
        overlap : float, optional
            A value between 0 and 1. If the area of two blobs overlaps by a
            fraction greater than `threshold`, the smaller blob is eliminated.
        log_scale : bool, optional
            If set intermediate values of standard deviations are interpolated
            using a logarithmic scale to the base `10`. If not, linear
            interpolation is used.
        exclude_border : int or bool, optional
            If nonzero int, `exclude_border` excludes blobs from
            within `exclude_border`-pixels of the border of the image.

        Returns
        -------
        A : (n, image.ndim + sigma) ndarray
            A 2d array with each row representing 2 coordinate values for a 2D
            image, and 3 coordinate values for a 3D image, plus the sigma(s) used.
            When a single sigma is passed, outputs are:
            ``(r, c, sigma)`` or ``(p, r, c, sigma)`` where ``(r, c)`` or
            ``(p, r, c)`` are coordinates of the blob and ``sigma`` is the standard
            deviation of the Gaussian kernel which detected the blob. When an
            anisotropic gaussian is used (sigmas per dimension), the detected sigma
            is returned for each dimension.

        References
        ----------
        .. [1] https://en.wikipedia.org/wiki/Blob_detection#The_Laplacian_of_Gaussian

        Examples
        --------
        >>> from skimage import data, feature, exposure
        >>> img = data.coins()
        >>> img = exposure.equalize_hist(img)  # improves detection
        >>> feature.blob_log(img, threshold = .3)
        array([[ 266.        ,  115.        ,   11.88888889],
            [ 263.        ,  302.        ,   17.33333333],
            [ 263.        ,  244.        ,   17.33333333],
            [ 260.        ,  174.        ,   17.33333333],
            [ 198.        ,  155.        ,   11.88888889],
            [ 198.        ,  103.        ,   11.88888889],
            [ 197.        ,   44.        ,   11.88888889],
            [ 194.        ,  276.        ,   17.33333333],
            [ 194.        ,  213.        ,   17.33333333],
            [ 185.        ,  344.        ,   17.33333333],
            [ 128.        ,  154.        ,   11.88888889],
            [ 127.        ,  102.        ,   11.88888889],
            [ 126.        ,  208.        ,   11.88888889],
            [ 126.        ,   46.        ,   11.88888889],
            [ 124.        ,  336.        ,   11.88888889],
            [ 121.        ,  272.        ,   17.33333333],
            [ 113.        ,  323.        ,    1.        ]])

        Notes
        -----
        The radius of each blob is approximately :math:`sqrt{2}sigma` for
        a 2-D image and :math:`sqrt{3}sigma` for a 3-D image.
        """
        image = img_as_float(image)

        # if both min and max sigma are scalar, function returns only one sigma
        scalar_sigma = (
            True if np.isscalar(max_sigma) and np.isscalar(min_sigma) else False
        )

        # Gaussian filter requires that sequence-type sigmas have same
        # dimensionality as image. This broadcasts scalar kernels
        if np.isscalar(max_sigma):
            max_sigma = np.full(image.ndim, max_sigma, dtype=float)
        if np.isscalar(min_sigma):
            min_sigma = np.full(image.ndim, min_sigma, dtype=float)

        # Convert sequence types to array
        min_sigma = np.asarray(min_sigma, dtype=float)
        max_sigma = np.asarray(max_sigma, dtype=float)

        if log_scale:
            start, stop = np.log10(min_sigma)[:, None], np.log10(max_sigma)[:, None]
            space = np.concatenate(
                [start, stop, np.full_like(start, num_sigma)], axis=1)
            sigma_list = np.stack([np.logspace(*s) for s in space], axis=1)
        else:
            scale = np.linspace(0, 1, num_sigma)[:, None]
            sigma_list = scale * (max_sigma - min_sigma) + min_sigma

        # computing gaussian laplace
        # average s**2 provides scale invariance
        gl_images = [-gaussian_laplace(image, s) * s ** 2
                     for s in np.mean(sigma_list, axis=1)]

        image_cube = np.stack(gl_images, axis=-1)

        local_maxima = peak_local_max(image_cube, threshold_abs=threshold,
                                      footprint=np.ones((3,) * (image.ndim + 1)),
                                      threshold_rel=0.0,
                                      exclude_border=exclude_border)

        # Catch no peaks
        if local_maxima.size == 0:
            return np.empty((0, 3))

        # Convert local_maxima to float64
        lm = local_maxima.astype(np.float64)

        # translate final column of lm, which contains the index of the
        # sigma that produced the maximum intensity value, into the sigma
        sigmas_of_peaks = sigma_list[local_maxima[:, -1]]

        if scalar_sigma:
            # select one sigma column, keeping dimension
            sigmas_of_peaks = sigmas_of_peaks[:, 0:1]

        # Remove sigma index and replace with sigmas
        lm = np.hstack([lm[:, :-1], sigmas_of_peaks])

        return _prune_blobs(lm, overlap)



import collections
import csv
import hashlib
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import (
    cast,
    FrozenSet,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Union)
from warnings import warn

import numpy as np
from slicedimage import ImageFormat, WriterContract

from starfish.core.types import Axes, Coordinates, CoordinateValue, Number
from . import TileIdentifier, write_irregular_experiment_json
from .inplace import InplaceFetchedTile, InplaceWriterContract
from .providers import FetchedTile, TileFetcher

FILENAME_CRE = re.compile(
    r"(?P<prefix>.+)-f(?P<fov>\d+)-r(?P<round>\d+)-c(?P<ch>\d+)-z(?P<zplane>\d+)\.(?P<ext>.+)")
TILE_COORDINATE_NAMES = ('fov', 'round', 'ch', 'zplane')


class ExtraPhysicalCoordinatesWarning(UserWarning):
    """Raised if there are physical coordinates in the coordinates csv file for a tile identifier
    that is not represented in the tile data.
    """
    pass


class PhysicalCoordinateNotPresentError(Exception):
    """Raised if physical coordinates along an axis is not present.  This is not a user-visible
    error because it may not actually be an error (i.e., exceptions of this class should be caught
    and handled in this module).
    """
    pass


@dataclass
class InferredTileResult:
    identifier: TileIdentifier
    path: Path
    format: ImageFormat


def format_structured_dataset(
        image_directory_path_str: str,
        physical_coordinates_csv_path_str: str,
        output_experiment_dir_str: str,
        tile_format: ImageFormat,
        in_place: bool = False,
) -> None:
    """
    Format a dataset where the filenames of the tiles provide most of the metadata required to
    organize them into the 7D tensor (image_type, fov, round, ch, zplane, y, x).  The remaining
    metadata, namely physical coordinates of the tiles, are provided in a CSV file.

    See the documentation in
    docs/source/_static/data_formatting_examples/exec_format_structured_data.py

    Parameters
    ----------
    image_directory_path_str : str
        The path of the directory where the image tiles are found.
    physical_coordinates_csv_path_str : str
        The path of the csv file containing the physical coordinates.
    output_experiment_dir_str : str
        The path of the directory to write the experiment to.  The top-level experiment json will be
        located at <output_experiment_dir_str>/experiment.json
    tile_format : ImageFormat
        The format of the tiles that are written.  Note that this must match the format of the
        original files if in_place is True.
    in_place : bool
        When this is true, the original tiles are used in the resulting experiment.json.
        (default: False)
    """
    stack_structure: Mapping[str, Sequence[InferredTileResult]] = \
        infer_stack_structure(Path(image_directory_path_str))
    physical_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = \
        read_physical_coordinates_from_csv(Path(physical_coordinates_csv_path_str))

    unseen_tile_coordinates = set(physical_coordinates.keys())

    # we should have physical coordinates for every region in the stack structure.
    for image_type, inferred_tile_results in stack_structure.items():
        for inferred_tile_result in inferred_tile_results:
            if inferred_tile_result.identifier not in physical_coordinates:
                raise ValueError(
                    f"Could not resolve physical coordinates for tile "
                    f"{inferred_tile_result.identifier} in image {image_type}"
                )
            if in_place and inferred_tile_result.format != tile_format:
                raise ValueError(
                    f"Tile {inferred_tile_result.identifier} has a different tile format ("
                    f"{inferred_tile_result.format}) than `format_structured_dataset` was called "
                    f"with ({tile_format}."
                )

            unseen_tile_coordinates.remove(inferred_tile_result.identifier)

    # do we have extra physical coordinates
    for tile_coordinate in unseen_tile_coordinates:
        warn(
            f"physical coordinates for tile {tile_coordinate} provided, but no tile exists at that"
            f"location.",
            ExtraPhysicalCoordinatesWarning,
        )

    # TODO: get a reasonable codebook imported...

    image_tile_identifiers: Mapping[str, Sequence[TileIdentifier]] = {
        image_type:
            [inferred_tile_result.identifier for inferred_tile_result in inferred_tile_results]
        for image_type, inferred_tile_results in stack_structure.items()
    }
    tile_fetchers: Mapping[str, TileFetcher] = {
        image_type: InferredTileFetcher(in_place, inferred_tile_results, physical_coordinates)
        for image_type, inferred_tile_results in stack_structure.items()
    }

    if in_place:
        writer_contract: Optional[WriterContract] = InplaceWriterContract()
    else:
        writer_contract = None

    write_irregular_experiment_json(
        output_experiment_dir_str,
        tile_format,
        image_tile_identifiers=image_tile_identifiers,
        tile_fetchers=tile_fetchers,
        writer_contract=writer_contract,
    )


def infer_stack_structure(
        basepath: Path,
) -> Mapping[str, Sequence[InferredTileResult]]:
    results: MutableMapping[str, MutableSequence[InferredTileResult]] = \
        collections.defaultdict(list)

    for path in basepath.glob("**/*"):
        mo = FILENAME_CRE.match(path.name)
        if mo is None:
            continue
        try:
            tile_format = ImageFormat.find_by_extension(mo.group('ext'))
        except ValueError:
            continue

        results[mo.group('prefix')].append(
            InferredTileResult(
                TileIdentifier(*[int(mo.group(component)) for component in TILE_COORDINATE_NAMES]),
                path,
                tile_format,
            )
        )

    return results


def read_physical_coordinates_from_csv(
        csvpath: Path,
) -> Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]]:
    required_coordinates: Mapping[Coordinates, str] = {
        Coordinates.X: "x",
        Coordinates.Y: "y",
    }
    coordinates_required_to_be_tuples: FrozenSet[Coordinates] = frozenset(
        (Coordinates.X, Coordinates.Y))

    result: MutableMapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = dict()
    with open(os.fspath(csvpath), "r") as raw_fh:
        for row_num, row in enumerate(csv.DictReader(raw_fh)):
            try:
                tile_coordinate = TileIdentifier(
                    *[int(row[component]) for component in TILE_COORDINATE_NAMES])
            except KeyError as ex:
                raise ValueError(
                    f"Could not find all the required columns on row {row_num + 1}") from ex

            physical_coordinates: MutableMapping[Coordinates, CoordinateValue] = {}
            for coordinate_name in list(Coordinates):
                try:
                    value = _parse_coordinates(row, coordinate_name)
                except PhysicalCoordinateNotPresentError:
                    # is this a required coordinate?
                    if coordinate_name in required_coordinates:
                        raise ValueError(
                            f"{required_coordinates[coordinate_name]} coordinates not found in row "
                            f"{row_num + 1}")
                    continue

                if (coordinate_name in coordinates_required_to_be_tuples
                        and not isinstance(value, tuple)):
                    raise ValueError(f"{coordinate_name.value} must be a range in row "
                                     f"{row_num + 1}")

                physical_coordinates[coordinate_name] = value

            result[tile_coordinate] = physical_coordinates

    return result


class InferredTile(InplaceFetchedTile):
    INPLACE_SHAPE: Optional[Mapping[Axes, int]] = None

    def __init__(
            self,
            in_place: bool,
            inferred_tile_result: InferredTileResult,
            tile_physical_coordinates:
            Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]],
    ) -> None:
        self.in_place = in_place
        self.inferred_tile_result = inferred_tile_result
        self.tile_physical_coordinates = tile_physical_coordinates
        self._tile_data: Optional[np.ndarray] = None

    @property
    def filepath(self) -> Path:
        return self.inferred_tile_result.path

    def _ensure_tile_loaded(self):
        if self._tile_data is None:
            if self.in_place and InferredTile.INPLACE_SHAPE is not None:
                self._tile_data = np.zeros(
                    (InferredTile.INPLACE_SHAPE[Axes.Y], InferredTile.INPLACE_SHAPE[Axes.X]),
                    dtype=np.float32,
                )
            else:
                self._tile_data = self.inferred_tile_result.format.reader_func(
                    os.fspath(self.inferred_tile_result.path))
                if self.in_place:
                    InferredTile.INPLACE_SHAPE = {
                        Axes.Y: self._tile_data.shape[0],
                        Axes.X: self._tile_data.shape[1],
                    }

    @property
    def sha256(self):
        with open(os.fspath(self.inferred_tile_result.path), "rb") as fh:
            return hashlib.sha256(fh.read()).hexdigest()

    @property
    def shape(self) -> Mapping[Axes, int]:
        tile_data = self.tile_data()
        return {Axes.Y: tile_data.shape[0], Axes.X: tile_data.shape[1]}

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]]:
        return self.tile_physical_coordinates

    def tile_data(self) -> np.ndarray:
        self._ensure_tile_loaded()
        return self._tile_data


class InferredTileFetcher(TileFetcher):
    def __init__(
            self,
            in_place: bool,
            inferred_tile_results: Sequence[InferredTileResult],
            physical_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]],
    ):
        self.in_place = in_place
        self.inferred_tile_results = inferred_tile_results
        self.physical_coordinates = physical_coordinates

    def get_tile(self, fov: int, r: int, ch: int, z: int) -> FetchedTile:
        wanted_tile_identifier = TileIdentifier(fov, r, ch, z)
        physical_coordinates = cast(
            Mapping[Union[str, Coordinates], Union[Number, Tuple[Number, Number]]],
            self.physical_coordinates[wanted_tile_identifier])

        for inferred_tile_result in self.inferred_tile_results:
            if wanted_tile_identifier == inferred_tile_result.identifier:
                return InferredTile(self.in_place, inferred_tile_result, physical_coordinates)

        raise KeyError(f"Could not find data for {wanted_tile_identifier}")


def _convert_str_to_Number(value_str: str) -> Number:
    """Converts a string into a Number.  Conversions to integers are preferred, and if that fails,
    we attempt to convert to a float."""
    try:
        return int(value_str)
    except ValueError:
        pass
    return float(value_str)


def _parse_coordinates(row, coordinate: Coordinates) -> CoordinateValue:
    """Given a row that may contain either no entries, a single scalar value, or a range for
    physical coordinates along an axis, attempt to return a CoordinateValue representing the input.
    In the case where no entries exist, :py:class:`PhysicalCoordinateNotPresentError` is raised.

    The columns will be denoted as the Coordinate value name, suffixed with `_min` or `_max`.  If
    both are present, then we interpret that as a range.  If only one is present, then we interpret
    that as a scalar value.
    """
    prefix = coordinate.value

    min_value = row.get(f"{prefix}_min", None)
    max_value = row.get(f"{prefix}_max", None)

    # convert empty or empty-ish strings into None.
    if min_value is not None and len(min_value.strip()) == 0:
        min_value = None
    if max_value is not None and len(max_value.strip()) == 0:
        max_value = None

    if min_value is not None and max_value is not None:
        return _convert_str_to_Number(min_value), _convert_str_to_Number(max_value)
    elif min_value is not None:
        return _convert_str_to_Number(min_value)
    elif max_value is not None:
        return _convert_str_to_Number(max_value)
    else:
        raise PhysicalCoordinateNotPresentError("neither present")



from typing import Dict, Hashable, Mapping, MutableMapping, Sequence, Tuple, Union

import numpy as np
import xarray as xr
from xarray.core.utils import is_scalar

from starfish.core.types import Axes, Coordinates, CoordinateValue


def convert_to_selector(
        indexers: Mapping[Axes, Union[int, slice, Sequence]]
) -> Mapping[Hashable, Union[int, slice, Sequence]]:
    """Converts a mapping of Axis to int, slice, or Sequence to a mapping of str to int or slice.
    The latter format is required for standard xarray indexing methods.

    Parameters
    ----------
    indexers : Mapping[Axes, Union[int, slice, Sequence]]
            A dictionary of dim:index where index is the value or range to index the dimension

    """
    return_dict: MutableMapping[Hashable, Union[int, slice, Sequence]] = {
        ind.value: ind.value if isinstance(ind.value, list)
        else slice(None, None) for ind in indexers.keys()}
    for key, value in indexers.items():
        if isinstance(value, tuple):
            return_dict[key.value] = slice(value[0], value[1])
        else:
            return_dict[key.value] = value
    return return_dict


def convert_coords_to_indices(
        array: xr.DataArray,
        indexers: Mapping[Coordinates, CoordinateValue],
) -> Dict[Axes, Union[int, Tuple[int, int]]]:
    """
    Convert mapping of physical coordinates to value or range to mapping of corresponding Axes and
    positional coordinates.

    Parameters
    ----------
    array : xr.DataArray
        The xarray with both physical and positional coordinates.
    indexers: Mapping[Coordinates, CoordinateValue]
        Mapping of physical coordinates to value or range

    Returns
    -------
    Mapping[Axes, Union[int, Tuple[int, int]]]:
        Mapping of Axes and positional indices that correspond to the given physical indices.

    """
    axes_indexers: Dict[Axes, Union[int, Tuple[int, int]]] = dict()
    if Coordinates.X in indexers:
        idx_x = find_nearest(array[Coordinates.X.value], indexers[Coordinates.X])
        axes_indexers[Axes.X] = idx_x
    if Coordinates.Y in indexers:
        idx_y = find_nearest(array[Coordinates.Y.value], indexers[Coordinates.Y])
        axes_indexers[Axes.Y] = idx_y
    if Coordinates.Z in indexers:
        idx_z = find_nearest(array[Coordinates.Z.value], indexers[Coordinates.Z])
        axes_indexers[Axes.ZPLANE] = idx_z
    return axes_indexers


def index_keep_dimensions(data: xr.DataArray,
                          indexers: Mapping[Hashable, Union[int, slice, Sequence]],
                          by_pos: bool=False
                          ) -> xr.DataArray:
    """Takes an xarray and key to index it. Indexes then adds back in lost dimensions"""
    # store original dims
    original_dims = data.dims
    # index
    if by_pos:
        data = data.isel(indexers)
    else:
        data = data.sel(indexers)
    # find missing dims
    missing_dims = set(original_dims) - set(data.dims)
    # Add back in missing dims
    data = data.expand_dims(tuple(missing_dims))

    # When the selection removes a dimension, xarray.expand_dims does not expand the non-indexed
    # dimensions that were removed.  For example, if one selects only a single zplane, it reduces
    # the z physical coordinate to a coordinate scalar, and not an array of size 1.  This hack
    # restores the dependent axes to arrays so they can be indexed.
    for primary_axis, dependent_axis in (
            (Axes.X, Coordinates.X),
            (Axes.Y, Coordinates.Y),
            (Axes.ZPLANE, Coordinates.Z),
    ):
        if primary_axis.value in missing_dims and is_scalar(data[dependent_axis.value]):
            data[dependent_axis.value] = xr.DataArray(
                np.array([data[dependent_axis.value]]),
                dims=primary_axis.value)

    # Reorder to correct format
    return data.transpose(*original_dims)


def find_nearest(array: xr.DataArray,
                 value: CoordinateValue
                 ) -> Union[int, Tuple[int, int]]:
    """
    Given an xarray and value or tuple range return the indices of the closest corresponding
    value/values in the array.

    Parameters
    ----------
    array: xr.DataArray
        The array to do lookups in.

    value : CoordinateValue
        The value or values to lookup.

    Returns
    -------
    Union[int, Tuple[int, int]]:
        The index or indicies of the entries closest to the given values in the array.
    """
    array = np.asarray(array)
    if isinstance(value, tuple):
        idx1 = (np.abs(array - value[0])).argmin()
        idx2 = (np.abs(array - value[1])).argmin()
        return idx1, idx2
    idx = (np.abs(array - value)).argmin()
    return idx



"""
Tests for IntensityTable.synthetic_intensities method.
"""

import numpy as np

from starfish.core.codebook.test.factories import codebook_array_factory
from starfish.core.types import Axes, Features
from ..intensity_table import IntensityTable


def test_synthetic_intensity_generation():
    """
    Create a 2-spot IntensityTable of pixel size (z=3, y=4, x=5) from a codebook with 3 channels
    and 2 rounds.

    Verify that the constructed Synthetic IntensityTable conforms to those dimensions, and given
    a known random seed, that the output spots decode to match a target in the input Codebook
    """
    # set seed to check that codebook is matched. This seed generates 2 instances of GENE_B
    np.random.seed(2)
    codebook = codebook_array_factory()
    num_z, height, width = 3, 4, 5
    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=num_z,
        height=height,
        width=width,
        n_spots=2,
    )

    # sizes should match codebook
    assert intensities.sizes[Axes.ROUND] == 2
    assert intensities.sizes[Axes.CH] == 3
    assert intensities.sizes[Features.AXIS] == 2

    # attributes should be bounded by the specified size
    assert np.all(intensities[Axes.ZPLANE.value] <= num_z)
    assert np.all(intensities[Axes.Y.value] <= height)
    assert np.all(intensities[Axes.X.value] <= width)

    # both codes should match GENE_B
    gene_b_intensities = codebook.sel(target="GENE_B")
    for feature_id in range(intensities.sizes[Features.AXIS]):
        feature_intensities = intensities[{Features.AXIS: feature_id}]
        assert np.array_equal(gene_b_intensities.values, feature_intensities.values)
